@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2009-2016 Yelp and Contributors
+# Copyright 2009-2018 Yelp and Contributors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -24,11 +24,13 @@ __credits__ = [
     'Jordan Andersen <jordandandersen@gmail.com>',
     'Tom Arnfeld <tarnfeld@me.com>',
     'Martin Baeuml <baeuml@kit.edu>',
+    'Benjamin Bariteau <bariteau@yelp.com>',
     'Étienne Batise <etienne.batise@gmail.com>',
     'Edward Betts <edward@4angle.com>',
     'Hunter Blanks <hblanks@monetate.com>',
     'Jim Blomo <jblomo@yelp.com>',
     'Reno Bowen <renobowen@gmail.com>',
+    'Stephen Brennan <sbrennan@yelp.com>',
     'James Brown <jbrown@uber.com>',
     'Kevin Burke <kevin@twilio.com>',
     'Tim Chan <macrotim@gmail.com>',
@@ -36,6 +38,7 @@ __credits__ = [
     'Ben Chess <bchess@yelp.com>',
     'Jordan Christensen <jc@kobo.com>',
     'Jonathan Chu <milki@yelp.com>',
+    'Ben Dalling <b.dalling@locp.co.uk>',
     'David Dehghan <ddehghan@gmail.com>',
     'Adam Derewecki <derewecki@gmail.com>',
     'Nick Dimiduk <ndimiduk@gmail.com>',
@@ -44,6 +47,8 @@ __credits__ = [
     'Tomer Elmalem <telmalem@gmail.com>',
     'Pedro Emanuel de Castro Faria Salgado <steenzout@ymail.com>',
     'Buck Evan <buck@yelp.com>',
+    'Aaron Friedlander <aaron.friedlander@affirm.com>',
+    'Cesar Gonzalez Fernandez <cesar.gon.fer@gmail.com>',
     'Dan Frank <danielhfrank@gmail.com>',
     'Sudarshan Gaikaiwari <sudarshan@acm.org>',
     'Alex Gaynor <alex.gaynor@gmail.com>',
@@ -69,13 +74,16 @@ __credits__ = [
     'Julian Krause <juliank@yelp.com>',
     'Pai-Wei Lai <paiwei@yelp.com>',
     'Boris Lau <boris.w.lau@gmail.com>',
+    'Laucia <laucia@users.noreply.github.com>',
     'Jerome Leclanche <jerome@leclan.ch>',
     'Robert Leftwich <rl.0x0@eml.cc>',
     'Andrew Lenards <andrew.lenards@gmail.com>',
+    'Ronny Lindner <ronny.lindner@invia.de>',
     'Tianhui Michael Li <tianhuil@cs.princeton.edu>',
     'James Lim <jim@affirm.com>',
     'Shusen Liu <liushusen.smart@gmail.com>',
     'Adrian Maceiras <amac425@utexas.edu>',
+    'Chet Mancini <chetmancini@gmail.com>',
     'Dallas Marlow <dallas@vineapp.com>',
     'Colin Matthias <cmatthias@cogolabs.com>',
     'Baris Metin <bmetin@yelp.com>',
@@ -94,6 +102,7 @@ __credits__ = [
     'Santeri Paavolainen <santtu@iki.fi>',
     'Matt Perry <matt@unshift.net>',
     'Kien Pham <kien@sendgrid.com>',
+    'Adam Phillabaum <adam@phillabaum.us>',
     'Andrew Price <andrew.price@ensighten.com>',
     'Wahbeh Qardaji <wahbeh.qardaji@gmail.com>',
     'Anusha Rajan <anusha@yelp.com>',
@@ -117,10 +126,13 @@ __credits__ = [
     'Matthew Tai <taim@google.com>',
     'Diogo Terror <me@diogoterror.com>',
     'William Vambenepe <vbp@google.com>',
+    'Daniel Vetter <lord_fright@yahoo.com>',
     'Paul Wais <pwais@yelp.com>',
+    'Qiu Wei <qiuwei.cw@alibaba-inc.com>',
     'Derek Wilson <jderekwilson@gmail.com>',
     'Tao Yu <taoyu@yelp.com>',
     'Andrea Zonca <andrea.zonca@gmail.com>',
+    'drulludanni <drulludanni5@gmail.com>',
 ]
 
-__version__ = '0.5.9'
+__version__ = '0.6.7'
