@@ -1,6 +1,7 @@
 # Copyright 2009-2012 Yelp
 # Copyright 2013 David Marin
-# Copyright 2015-2016 Yelp
+# Copyright 2015-2017 Yelp
+# Copyright 2018 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -33,105 +34,9 @@ from mrjob.py2 import string_types
 from mrjob.util import expand_path
 from mrjob.util import shlex_split
 
-
 log = logging.getLogger(__name__)
 
 
-class OptionStore(dict):
-    """Encapsulates logic about a configuration. With the exception of the
-    constructor, it can be accessed like a dictionary."""
-
-    #: Set of valid keys for this type of configuration
-    ALLOWED_KEYS = set()
-
-    #: Mapping of key to function used to combine multiple values to override,
-    #: augment, etc. Leave blank for :py:func:`combine_values()`.
-    COMBINERS = {}
-
-    #: Mapping from old name for an option to its new name
-    DEPRECATED_ALIASES = {}
-
-    def __init__(self):
-        super(OptionStore, self).__init__()
-        self.cascading_dicts = [
-            dict((key, None) for key in self.ALLOWED_KEYS),
-            self.default_options(),
-        ]
-
-    def default_options(self):
-        """Default options for this :py:class:`OptionStore`"""
-        return {}
-
-    def validated_options(self, opts, from_where=''):
-        results = {}
-
-        for k, v in sorted(opts.items()):
-            if k in self.DEPRECATED_ALIASES:
-                if v is None:
-                    continue
-
-                aliased_opt = self.DEPRECATED_ALIASES[k]
-
-                log.warning('Deprecated option %s%s has been renamed to %s'
-                            ' and will be removed in v0.6.0' % (
-                                k, from_where, aliased_opt))
-
-                if opts.get(aliased_opt) is None:
-                    results[aliased_opt] = v
-
-            elif k in self.ALLOWED_KEYS:
-                results[k] = v
-
-            else:
-                log.warning('Unexpected option %s%s' % (k, from_where))
-
-        return results
-
-    def populate_values_from_cascading_dicts(self):
-        """When ``cascading_dicts`` has been built, use it to populate the
-        dictionary with the ultimate values.
-        """
-        self.update(combine_opts(self.COMBINERS, *self.cascading_dicts))
-        self._opt_priority = self._calculate_opt_priority()
-
-    def is_default(self, key):
-        return self._opt_priority[key] < 2
-
-    def __getitem__(self, key):
-        if key in self.ALLOWED_KEYS:
-            return super(OptionStore, self).__getitem__(key)
-        else:
-            raise KeyError(key)
-
-    def __setitem__(self, key, value):
-        if key in self.ALLOWED_KEYS:
-            return super(OptionStore, self).__setitem__(key, value)
-        else:
-            raise KeyError(key)
-
-    def _calculate_opt_priority(self):
-        """Keep track of where in the order opts were specified,
-        to handle opts that affect the same thing (e.g. ec2_*instance_type).
-
-        Here is a rough guide to the values set by this function. They are
-
-            Where specified     Priority
-            unset everywhere    -1
-            blank               0
-            non-blank default   1
-            base conf file      2
-            inheriting conf     [3-n]
-            command line        n+1
-        """
-        opt_priority = dict((opt, -1) for opt in self)
-        for priority, opt_dict in enumerate(self.cascading_dicts):
-            if opt_dict:
-                for opt, value in opt_dict.items():
-                    if value is not None:
-                        opt_priority[opt] = priority
-        return opt_priority
-
-
 ### finding config files ###
 
 def find_mrjob_conf():
@@ -346,13 +251,11 @@ def load_opts_from_mrjob_conf(runner_alias, conf_path=None,
                            been loaded (used by
                            :py:func:`load_opts_from_mrjob_confs`).
 
-    .. versionchanged:: 0.4.6
-        Relative ``include:`` paths are relative to the real (after resolving
-        symlinks) path of the including conf file
+    Relative ``include:`` paths are relative to the real (after resolving
+    symlinks) path of the including conf file
 
-    .. versionchanged:: 0.4.6
-        This will only load each config file once, even if it's referenced
-        from multiple paths due to symlinks.
+    This will only load each config file once, even if it's referenced
+    from multiple paths due to symlinks.
     """
     if already_loaded is None:
         already_loaded = []
@@ -419,12 +322,11 @@ def load_opts_from_mrjob_confs(runner_alias, conf_paths=None):
     :type conf_paths: list or ``None``
     :param conf_path: locations of the files to load
 
-    .. versionchanged:: 0.4.6
-        This will only load each config file once, even if it's referenced
-        from multiple paths due to symlinks.
+    This will only load each config file once, even if it's referenced
+    from multiple paths due to symlinks.
     """
     if conf_paths is None:
-        return load_opts_from_mrjob_conf(runner_alias)
+        results = load_opts_from_mrjob_conf(runner_alias)
     else:
         # don't include conf files that were loaded earlier in conf_paths
         already_loaded = []
@@ -437,7 +339,10 @@ def load_opts_from_mrjob_confs(runner_alias, conf_paths=None):
             results = load_opts_from_mrjob_conf(
                 runner_alias, path, already_loaded=already_loaded) + results
 
-        return results
+    if runner_alias and not any(conf for path, conf in results):
+        log.warning('No configs specified for %s runner' % runner_alias)
+
+    return results
 
 
 ### writing mrjob.conf ###
@@ -492,9 +397,8 @@ def combine_lists(*seqs):
     Generally this is used for a list of commands we want to run; the
     "default" commands get run before any commands specific to your job.
 
-    .. versionchanged:: 0.4.6
-       Strings, bytes, and non-sequence objects (e.g. numbers) are treated as
-       single-item lists.
+    Strings, bytes, and non-sequence objects (e.g. numbers) are treated as
+    single-item lists.
 
     .. versionchanged:: 0.5.3
        Dicts are treated as single-item lists
@@ -610,6 +514,15 @@ def _combine_envs_helper(envs, local):
     return result
 
 
+def combine_jobconfs(*jobconfs):
+    """Like combine_dicts(), but non-string values are converted to
+    Java-readable string (e.g. True becomes 'true'). Keys whose
+    value is None are blanked out."""
+    j = combine_dicts(*jobconfs)
+
+    return {k: _to_java_str(v) for k, v in j.items() if v is not None}
+
+
 def combine_paths(*paths):
     """Returns the last value in *paths* that is not ``None``.
     Resolve ``~`` (home dir) and environment variables."""
@@ -621,8 +534,7 @@ def combine_path_lists(*path_seqs):
     Resolve ``~`` (home dir) and environment variables, and expand globs
     that refer to the local filesystem.
 
-    .. versionchanged:: 0.4.6
-       Can take single strings as well as lists.
+    Can take single strings as well as lists.
     """
     results = []
 
@@ -667,3 +579,19 @@ def combine_opts(combiners, *opts_list):
         final_opts[key] = combine_func(*values)
 
     return final_opts
+
+
+def _to_java_str(x):
+    """Convert a value (usually for a configuration property) into its
+    Java string representation, falling back to the Python representation
+    if None is available."""
+    # e.g. True -> 'true', None -> 'null'. See #323
+    if isinstance(x, string_types):
+        return x
+    elif x is None:
+        # Note: combine_jobconfs() blanks out keys with None values
+        return 'null'
+    elif isinstance(x, bool):
+        return 'true' if x else 'false'
+    else:
+        return str(x)
