@@ -1,4 +1,6 @@
 # Copyright 2009-2016 Yelp and Contributors
+# Copyright 2017 Yelp
+# Copyright 2018 Yelp and Google, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -26,26 +28,21 @@ try:
 except ImportError:
     pty = None
 
-from mrjob.compat import translate_jobconf
+from mrjob.bin import MRJobBinRunner
 from mrjob.compat import uses_yarn
 from mrjob.conf import combine_dicts
 from mrjob.fs.composite import CompositeFilesystem
 from mrjob.fs.hadoop import HadoopFilesystem
 from mrjob.fs.local import LocalFilesystem
-from mrjob.logs.counters import _format_counters
 from mrjob.logs.counters import _pick_counters
 from mrjob.logs.errors import _format_error
 from mrjob.logs.mixin import LogInterpretationMixin
 from mrjob.logs.step import _interpret_hadoop_jar_command_stderr
 from mrjob.logs.step import _is_counter_log4j_record
 from mrjob.logs.wrap import _logs_exist
-from mrjob.options import _allowed_keys
-from mrjob.options import _combiners
-from mrjob.options import _deprecated_aliases
 from mrjob.parse import is_uri
-from mrjob.py2 import to_string
-from mrjob.runner import MRJobRunner
-from mrjob.runner import RunnerOptionStore
+from mrjob.py2 import to_unicode
+from mrjob.runner import _fix_env
 from mrjob.setup import UploadDirManager
 from mrjob.step import StepFailedException
 from mrjob.step import _is_spark_step_type
@@ -83,14 +80,14 @@ _FALLBACK_HADOOP_YARN_LOG_DIRS = [
 ]
 
 # start of Counters printed by Hadoop
-_HADOOP_COUNTERS_START_RE = re.compile(b'^Counters: (?P<amount>\d+)\s*$')
+_HADOOP_COUNTERS_START_RE = re.compile(br'^Counters: (?P<amount>\d+)\s*$')
 
 # header for a group of counters
-_HADOOP_COUNTER_GROUP_RE = re.compile(b'^(?P<indent>\s+)(?P<group>.*)$')
+_HADOOP_COUNTER_GROUP_RE = re.compile(br'^(?P<indent>\s+)(?P<group>.*)$')
 
 # line for a counter
 _HADOOP_COUNTER_RE = re.compile(
-    b'^(?P<indent>\s+)(?P<counter>.*)=(?P<amount>\d+)\s*$')
+    br'^(?P<indent>\s+)(?P<counter>.*)=(?P<amount>\d+)\s*$')
 
 # the one thing Hadoop streaming prints to stderr not in log format
 _HADOOP_NON_LOG_LINE_RE = re.compile(r'^Streaming Command Failed!')
@@ -114,21 +111,7 @@ def fully_qualify_hdfs_path(path):
         return 'hdfs:///user/%s/%s' % (getpass.getuser(), path)
 
 
-class HadoopRunnerOptionStore(RunnerOptionStore):
-
-    ALLOWED_KEYS = _allowed_keys('hadoop')
-    COMBINERS = _combiners('hadoop')
-    DEPRECATED_ALIASES = _deprecated_aliases('hadoop')
-
-    def default_options(self):
-        super_opts = super(HadoopRunnerOptionStore, self).default_options()
-        return combine_dicts(super_opts, {
-            'hadoop_tmp_dir': 'tmp/mrjob',
-            'spark_master': 'yarn',
-        })
-
-
-class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
+class HadoopJobRunner(MRJobBinRunner, LogInterpretationMixin):
     """Runs an :py:class:`~mrjob.job.MRJob` on your Hadoop cluster.
     Invoked when you run your job with ``-r hadoop``.
 
@@ -137,7 +120,19 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
     """
     alias = 'hadoop'
 
-    OPTION_STORE_CLASS = HadoopRunnerOptionStore
+    OPT_NAMES = MRJobBinRunner.OPT_NAMES | {
+        'hadoop_bin',
+        'hadoop_extra_args',
+        'hadoop_log_dirs',
+        'hadoop_streaming_jar',
+        'hadoop_tmp_dir',
+        'spark_deploy_mode',
+        'spark_master',
+    }
+
+    # supports everything (so far)
+    _STEP_TYPES = {
+        'jar', 'spark', 'spark_jar', 'spark_script', 'streaming'}
 
     def __init__(self, **kwargs):
         """:py:class:`~mrjob.hadoop.HadoopJobRunner` takes the same arguments
@@ -146,13 +141,6 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
         """
         super(HadoopJobRunner, self).__init__(**kwargs)
 
-        if self._opts['hadoop_home']:
-            log.warning(
-                'hadoop_home is deprecated since 0.5.0 and will be removed'
-                ' in v0.6.0. In most cases, mrjob will now find the hadoop'
-                ' binary and streaming jar without help. If not, use the'
-                ' hadoop_bin and hadoop_streaming_jar options.')
-
         self._hadoop_tmp_dir = fully_qualify_hdfs_path(
             posixpath.join(
                 self._opts['hadoop_tmp_dir'], self._job_key))
@@ -189,6 +177,16 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
         # others will be filled as needed)
         self._log_interpretations = []
 
+    def _default_opts(self):
+        return combine_dicts(
+            super(HadoopJobRunner, self)._default_opts(),
+            dict(
+                hadoop_tmp_dir='tmp/mrjob',
+                spark_deploy_mode='client',
+                spark_master='yarn',
+            )
+        )
+
     @property
     def fs(self):
         """:py:class:`mrjob.fs.base.Filesystem` object for HDFS and the local
@@ -253,9 +251,6 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
     def _hadoop_dirs(self):
         """Yield all possible hadoop directories (used for streaming jar
         and logs). May yield duplicates"""
-        if self._opts['hadoop_home']:
-            yield self._opts['hadoop_home']
-
         for name in ('HADOOP_PREFIX', 'HADOOP_HOME', 'HADOOP_INSTALL',
                      'HADOOP_MAPRED_HOME'):
             path = os.environ.get(name)
@@ -354,8 +349,7 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
 
     def _run(self):
         self._find_binaries_and_jars()
-        self._check_input_exists()
-        self._create_setup_wrapper_script()
+        self._create_setup_wrapper_scripts()
         self._add_job_files_for_upload()
         self._upload_local_files_to_hdfs()
         self._run_job_in_hadoop()
@@ -376,26 +370,13 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
         if self._has_spark_steps():
             self.get_spark_submit_bin()
 
-    def _check_input_exists(self):
-        """Make sure all input exists before continuing with our job.
-        """
-        for path in self._input_paths:
-            if path == '-':
-                continue  # STDIN always exists
-
-            if self._opts['check_input_paths']:
-                if not self.fs.exists(path):
-                    raise AssertionError(
-                        'Input path %s does not exist!' % (path,))
-
     def _add_job_files_for_upload(self):
         """Add files needed for running the job (setup and input)
         to self._upload_mgr."""
-        for path in self._get_input_paths():
-            self._upload_mgr.add(path)
-
         for path in self._working_dir_mgr.paths():
             self._upload_mgr.add(path)
+        for path in self._py_files():
+            self._upload_mgr.add(path)
 
     def _upload_local_files_to_hdfs(self):
         """Copy files managed by self._upload_mgr to HDFS
@@ -428,7 +409,7 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
             self._warn_about_spark_archives(step)
 
             step_args = self._args_for_step(step_num)
-            env = self._env_for_step(step_num)
+            env = _fix_env(self._env_for_step(step_num))
 
             # log this *after* _args_for_step(), which can start a search
             # for the Hadoop streaming jar
@@ -458,7 +439,7 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
 
                 # there shouldn't be much output to STDOUT
                 for line in step_proc.stdout:
-                    _log_line_from_hadoop(to_string(line).strip('\r\n'))
+                    _log_line_from_hadoop(to_unicode(line).strip('\r\n'))
 
                 step_proc.stdout.close()
                 step_proc.stderr.close()
@@ -487,14 +468,9 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
 
             log_interpretation['step'] = step_interpretation
 
-            step_type = step['type']
+            self._log_counters(log_interpretation, step_num)
 
-            if not _is_spark_step_type(step_type):
-                counters = self._pick_counters(log_interpretation, step_type)
-                if counters:
-                    log.info(_format_counters(counters))
-                else:
-                    log.warning('No counters found')
+            step_type = step['type']
 
             if returncode:
                 error = self._pick_error(log_interpretation, step_type)
@@ -528,55 +504,15 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
         elif _is_spark_step_type(step['type']):
             return self._args_for_spark_step(step_num)
         else:
-            raise AssertionError('Bad step type: %r' % (step['type'],))
+            raise ValueError('Bad step type: %r' % (step['type'],))
 
     def _args_for_streaming_step(self, step_num):
         hadoop_streaming_jar = self.get_hadoop_streaming_jar()
         if not hadoop_streaming_jar:
             raise Exception('no Hadoop streaming jar')
 
-        mapper, combiner, reducer = (
-            self._hadoop_streaming_commands(step_num))
-
-        args = self.get_hadoop_bin() + ['jar', hadoop_streaming_jar]
-
-        # set up uploading from HDFS to the working dir
-        args.extend(self._upload_args())
-
-        # if no reducer, shut off reducer tasks. This has to come before
-        # extra hadoop args, which could contain jar-specific args
-        # (e.g. -outputformat). See #1331.
-        #
-        # might want to just integrate this into _hadoop_args_for_step?
-        if not reducer:
-            args.extend(['-D', ('%s=0' % translate_jobconf(
-                'mapreduce.job.reduces', self.get_hadoop_version()))])
-
-        # Add extra hadoop args first as hadoop args could be a hadoop
-        # specific argument which must come before job
-        # specific args.
-        args.extend(self._hadoop_args_for_step(step_num))
-
-        # set up input
-        for input_uri in self._step_input_uris(step_num):
-            args.extend(['-input', input_uri])
-
-        # set up output
-        args.append('-output')
-        args.append(self._step_output_uri(step_num))
-
-        args.append('-mapper')
-        args.append(mapper)
-
-        if combiner:
-            args.append('-combiner')
-            args.append(combiner)
-
-        if reducer:
-            args.append('-reducer')
-            args.append(reducer)
-
-        return args
+        return (self.get_hadoop_bin() + ['jar', hadoop_streaming_jar] +
+                self._hadoop_streaming_jar_args(step_num))
 
     def _args_for_jar_step(self, step_num):
         step = self._get_step(step_num)
@@ -585,9 +521,6 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
 
         args.extend(self.get_hadoop_bin())
 
-        # -libjars, -D
-        args.extend(self._hadoop_generic_args_for_step(step_num))
-
         # special case for consistency with EMR runner.
         #
         # This might look less like duplicated code if we ever
@@ -604,13 +537,10 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
 
         if step.get('args'):
             args.extend(
-                self._interpolate_input_and_output(step['args'], step_num))
+                self._interpolate_step_args(step['args'], step_num))
 
         return args
 
-    def _spark_submit_arg_prefix(self):
-        return ['--master', self._opts['spark_master']]
-
     def _env_for_step(self, step_num):
         step = self._get_step(step_num)
 
@@ -634,10 +564,20 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
             except Exception as e:
                 log.exception(e)
 
+    def _manifest_download_commands(self):
+        cp_to_local = self.get_hadoop_bin() + ['fs', '-copyToLocal']
+
+        return [
+            ('*://*', cmd_line(cp_to_local)),
+        ]
+
     ### LOG (implementation of LogInterpretationMixin) ###
 
     def _stream_history_log_dirs(self, output_dir=None):
         """Yield lists of directories to look for the history log in."""
+        if not self._read_logs():
+            return
+
         for log_dir in unique(self._hadoop_log_dirs(output_dir=output_dir)):
             if _logs_exist(self.fs, log_dir):
                 log.info('Looking for history log in %s...' % log_dir)
@@ -649,6 +589,9 @@ class HadoopJobRunner(MRJobRunner, LogInterpretationMixin):
         # Note: this is unlikely to be super-helpful on "real" (multi-node)
         # pre-YARN Hadoop because task logs aren't generally shipped to a
         # local directory. It's a start, anyways. See #1201.
+        if not self._read_logs():
+            return
+
         for log_dir in unique(self._hadoop_log_dirs(output_dir=output_dir)):
             if application_id:
                 path = self.fs.join(log_dir, 'userlogs', application_id)
