@@ -1,4 +1,5 @@
 # Copyright 2009-2017 Yelp and Contributors
+# Copyright 2018 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,20 +17,20 @@ for more information."""
 
 # don't add imports here that aren't part of the standard Python library,
 # since MRJobs need to run in Amazon's generic EMR environment
+import codecs
 import inspect
 import itertools
 import json
 import logging
 import os.path
 import sys
-from optparse import OptionGroup
 
 # don't use relative imports, to allow this script to be invoked as __main__
-from mrjob.conf import combine_dicts
-from mrjob.conf import combine_lists
+from mrjob.cat import decompress
 from mrjob.launch import MRJobLauncher
+from mrjob.launch import _im_func
 from mrjob.launch import _READ_ARGS_FROM_SYS_ARGV
-from mrjob.options import _add_step_options
+from mrjob.options import _add_step_args
 from mrjob.options import _print_help_for_steps
 from mrjob.protocol import JSONProtocol
 from mrjob.protocol import RawValueProtocol
@@ -39,26 +40,12 @@ from mrjob.step import MRStep
 from mrjob.step import SparkStep
 from mrjob.step import _JOB_STEP_FUNC_PARAMS
 from mrjob.util import expand_path
-from mrjob.util import read_input
+from mrjob.util import to_lines
 
 
 log = logging.getLogger(__name__)
 
 
-def _im_func(f):
-    """Wrapper to get at the underlying function belonging to a method.
-
-    Python 2 is slightly different because classes have "unbound methods"
-    which wrap the underlying function, whereas on Python 3 they're just
-    functions. (Methods work the same way on both versions.)
-    """
-    # "im_func" is the old Python 2 name for __func__
-    if hasattr(f, '__func__'):
-        return f.__func__
-    else:
-        return f
-
-
 class UsageError(Exception):
     pass
 
@@ -67,6 +54,9 @@ class MRJob(MRJobLauncher):
     """The base class for all MapReduce jobs. See :py:meth:`__init__`
     for details."""
 
+    # script path is whatever file our subclass of MRJob is in
+    _FIRST_ARG_IS_SCRIPT_PATH = False
+
     def __init__(self, args=None):
         """Entry point for running your job from other Python code.
 
@@ -86,9 +76,11 @@ class MRJob(MRJobLauncher):
         """
         super(MRJob, self).__init__(self.mr_job_script(), args)
 
+        self._warned_about_parse_output_line = False
+
     @classmethod
     def _usage(cls):
-        return "usage: %prog [options] [input files]"
+        return "%(prog)s [options] [input files]"
 
     ### Defining one-step streaming jobs ###
 
@@ -197,6 +189,17 @@ class MRJob(MRJobLauncher):
         """
         raise NotImplementedError
 
+    def mapper_raw(self, input_path, input_uri):
+        """Re-define this to make Hadoop pass one input file to each
+        mapper.
+
+        :param input_path: a local path that the input file has been copied to
+        :param input_uri: the URI of the input file on HDFS, S3, etc
+
+        .. versionadded:: 0.6.3
+        """
+        raise NotImplementedError
+
     def reducer_init(self):
         """Re-define this to define an action to run before the reducer
         processes any input.
@@ -321,12 +324,14 @@ class MRJob(MRJobLauncher):
         re-defined. For example::
 
             def steps(self):
-                return [self.mr(mapper=self.transform_input,
-                                reducer=self.consolidate_1),
-                        self.mr(reducer_init=self.log_mapper_init,
-                                reducer=self.consolidate_2)]
+                return [MRStep(mapper=self.transform_input,
+                               reducer=self.consolidate_1),
+                        MRStep(reducer_init=self.log_mapper_init,
+                               reducer=self.consolidate_2)]
 
-        :return: a list of steps constructed with :py:meth:`mr`
+        :return: a list of steps constructed with
+                 :py:class:`~mrjob.step.MRStep` or other classes in
+                 :py:mod:`mrjob.step`.
         """
         # only include methods that have been redefined
         kwargs = dict(
@@ -354,16 +359,10 @@ class MRJob(MRJobLauncher):
 
         kwargs.update(updates)
 
-        return [MRStep(**kwargs)]
-
-    @classmethod
-    def mr(cls, **kwargs):
-        """A deprecated wrapper for :py:class:`~mrjob.step.MRStep`.
-        """
-        log.warning('mr() is deprecated and will be removed in v0.6.0.'
-                    ' Use mrjob.step.MRStep directly instead.')
-
-        return MRStep(**kwargs)
+        if kwargs:
+            return [MRStep(**kwargs)]
+        else:
+            return []
 
     def increment_counter(self, group, counter, amount=1):
         """Increment a counter in Hadoop streaming by printing to stderr.
@@ -439,9 +438,15 @@ class MRJob(MRJobLauncher):
         mr_job.execute()
 
     def execute(self):
-        # MRJob does Hadoop Streaming stuff, or defers to Launcher (superclass)
-        # if not otherwise instructed
+        # MRJob does Hadoop Streaming stuff, or defers to its superclass
+        # (MRJobLauncher) if not otherwise instructed
         if self.options.show_steps:
+            log_stream = codecs.getwriter('utf_8')(self.stderr)
+
+            self.set_up_logging(quiet=self.options.quiet,
+                                verbose=self.options.verbose,
+                                stream=log_stream)
+
             self.show_steps()
 
         elif self.options.run_mapper:
@@ -474,16 +479,30 @@ class MRJob(MRJobLauncher):
                                  " probably means you tried to use it from"
                                  " __main__, which doesn't work." % w)
 
-        # support inline runner when running from the MRJob itself
-        from mrjob.inline import InlineMRJobRunner
-
-        # inline is the default, not local
-        if not self.options.runner or self.options.runner == 'inline':
-            return InlineMRJobRunner(mrjob_cls=self.__class__,
-                                     **self.inline_job_runner_kwargs())
-
         return super(MRJob, self).make_runner()
 
+    def _runner_class(self):
+        """Runner class as indicated by ``--runner``. Defaults to ``'inline'``.
+        """
+        if not self.options.runner or self.options.runner == 'inline':
+            from mrjob.inline import InlineMRJobRunner
+            return InlineMRJobRunner
+
+        else:
+            return super(MRJob, self)._runner_class()
+
+    def _runner_kwargs(self):
+        """If we're building an inline runner, include mrjob_cls in kwargs."""
+        kwargs = super(MRJob, self)._runner_kwargs()
+
+        if self._runner_class().alias == 'inline':
+            kwargs = dict(mrjob_cls=self.__class__, **kwargs)
+
+        # pass steps to runner (see #1845)
+        kwargs = dict(steps=self._steps_desc(), **kwargs)
+
+        return kwargs
+
     def _get_step(self, step_num, expected_type):
         """Helper for run_* methods"""
         steps = self.steps()
@@ -500,78 +519,14 @@ class MRJob(MRJobLauncher):
         :type step_num: int
         :param step_num: which step to run (0-indexed)
 
-        If --no-strict-protocols is set, and we encounter a line that can't
-        be decoded by our input protocol or a tuple that can't be encoded
-        by our output protocol, we'll increment a counter rather than
-        raising an exception. This will be going away in v0.6.0.
-
         Called from :py:meth:`run`. You'd probably only want to call this
         directly from automated tests.
         """
-        step = self._get_step(step_num, MRStep)
-
-        mapper = step['mapper']
-        mapper_init = step['mapper_init']
-        mapper_final = step['mapper_final']
-
         # pick input and output protocol
         read_lines, write_line = self._wrap_protocols(step_num, 'mapper')
 
-        if mapper_init:
-            for out_key, out_value in mapper_init() or ():
-                write_line(out_key, out_value)
-
-        # run the mapper on each line
-        for key, value in read_lines():
-            for out_key, out_value in mapper(key, value) or ():
-                write_line(out_key, out_value)
-
-        if mapper_final:
-            for out_key, out_value in mapper_final() or ():
-                write_line(out_key, out_value)
-
-    def run_reducer(self, step_num=0):
-        """Run the reducer for the given step.
-
-        :type step_num: int
-        :param step_num: which step to run (0-indexed)
-
-        If we encounter a line that can't be decoded by our input protocol,
-        or a tuple that can't be encoded by our output protocol, we'll
-        increment a counter rather than raising an exception. If
-        --strict-protocols is set, then an exception is raised
-
-        Called from :py:meth:`run`. You'd probably only want to call this
-        directly from automated tests.
-        """
-        step = self._get_step(step_num, MRStep)
-
-        reducer = step['reducer']
-        reducer_init = step['reducer_init']
-        reducer_final = step['reducer_final']
-        if reducer is None:
-            raise ValueError('No reducer in step %d' % step_num)
-
-        # pick input and output protocol
-        read_lines, write_line = self._wrap_protocols(step_num, 'reducer')
-
-        if reducer_init:
-            for out_key, out_value in reducer_init() or ():
-                write_line(out_key, out_value)
-
-        # group all values of the same key together, and pass to the reducer
-        #
-        # be careful to use generators for everything, to allow for
-        # very large groupings of values
-        for key, kv_pairs in itertools.groupby(read_lines(),
-                                               key=lambda k_v: k_v[0]):
-            values = (v for k, v in kv_pairs)
-            for out_key, out_value in reducer(key, values) or ():
-                write_line(out_key, out_value)
-
-        if reducer_final:
-            for out_key, out_value in reducer_final() or ():
-                write_line(out_key, out_value)
+        for k, v in self.map_pairs(read_lines(), step_num=step_num):
+            write_line(k, v)
 
     def run_combiner(self, step_num=0):
         """Run the combiner for the given step.
@@ -587,34 +542,124 @@ class MRJob(MRJobLauncher):
         Called from :py:meth:`run`. You'd probably only want to call this
         directly from automated tests.
         """
-        step = self._get_step(step_num, MRStep)
-
-        combiner = step['combiner']
-        combiner_init = step['combiner_init']
-        combiner_final = step['combiner_final']
-        if combiner is None:
-            raise ValueError('No combiner in step %d' % step_num)
-
         # pick input and output protocol
         read_lines, write_line = self._wrap_protocols(step_num, 'combiner')
 
-        if combiner_init:
-            for out_key, out_value in combiner_init() or ():
-                write_line(out_key, out_value)
+        for k, v in self.combine_pairs(read_lines(), step_num=step_num):
+            write_line(k, v)
 
-        # group all values of the same key together, and pass to the combiner
+    def run_reducer(self, step_num=0):
+        """Run the reducer for the given step.
+
+        :type step_num: int
+        :param step_num: which step to run (0-indexed)
+
+        Called from :py:meth:`run`. You'd probably only want to call this
+        directly from automated tests.
+        """
+        # pick input and output protocol
+        read_lines, write_line = self._wrap_protocols(step_num, 'reducer')
+
+        for k, v in self.reduce_pairs(read_lines(), step_num=step_num):
+            write_line(k, v)
+
+    def map_pairs(self, pairs, step_num=0):
+        """Runs :py:meth:`mapper_init`,
+        :py:meth:`mapper`/:py:meth:`mapper_raw`, and :py:meth:`mapper_final`
+        for one map task in one step.
+
+        Takes in a sequence of (key, value) pairs as input, and yields
+        (key, value) pairs as output.
+
+        :py:meth:`run_mapper` essentially wraps this method with code to handle
+        reading/decoding input and writing/encoding output.
+
+        .. versionadded:: 0.6.7
+        """
+        step = self._get_step(step_num, MRStep)
+
+        mapper = step['mapper']
+        mapper_raw = step['mapper_raw']
+        mapper_init = step['mapper_init']
+        mapper_final = step['mapper_final']
+
+        if mapper_init:
+            for k, v in mapper_init() or ():
+                yield k, v
+
+        if mapper_raw:
+            if len(self.options.args) != 2:
+                raise ValueError('Wrong number of args')
+            input_path, input_uri = self.options.args
+            for k, v in mapper_raw(input_path, input_uri) or ():
+                yield k, v
+        else:
+            for key, value in pairs:
+                for k, v in mapper(key, value) or ():
+                    yield k, v
+
+        if mapper_final:
+            for k, v in mapper_final() or ():
+                yield k, v
+
+    def combine_pairs(self, pairs, step_num=0):
+        """Runs :py:meth:`combiner_init`,
+        :py:meth:`combiner`, and :py:meth:`combiner_final`
+        for one reduce task in one step.
+
+        Takes in a sequence of (key, value) pairs as input, and yields
+        (key, value) pairs as output.
+
+        :py:meth:`run_combiner` essentially wraps this method with code to
+        handle reading/decoding input and writing/encoding output.
+
+        .. versionadded:: 0.6.7
+        """
+        for k, v in self._combine_or_reduce_pairs(pairs, 'combiner', step_num):
+            yield k, v
+
+    def reduce_pairs(self, pairs, step_num=0):
+        """Runs :py:meth:`reducer_init`,
+        :py:meth:`reducer`, and :py:meth:`reducer_final`
+        for one reduce task in one step.
+
+        Takes in a sequence of (key, value) pairs as input, and yields
+        (key, value) pairs as output.
+
+        :py:meth:`run_reducer` essentially wraps this method with code to
+        handle reading/decoding input and writing/encoding output.
+
+        .. versionadded:: 0.6.7
+        """
+        for k, v in self._combine_or_reduce_pairs(pairs, 'reducer', step_num):
+            yield k, v
+
+    def _combine_or_reduce_pairs(self, pairs, mrc, step_num=0):
+        """Helper for :py:meth:`combine_pairs` and :py:meth:`reduce_pairs`."""
+        step = self._get_step(step_num, MRStep)
+
+        task = step[mrc]
+        task_init = step[mrc + '_init']
+        task_final = step[mrc + '_final']
+        if task is None:
+            raise ValueError('No %s in step %d' % (mrc, step_num))
+
+        if task_init:
+            for k, v in task_init() or ():
+                yield k, v
+
+        # group all values of the same key together, and pass to the reducer
         #
         # be careful to use generators for everything, to allow for
         # very large groupings of values
-        for key, kv_pairs in itertools.groupby(read_lines(),
-                                               key=lambda k_v1: k_v1[0]):
-            values = (v for k, v in kv_pairs)
-            for out_key, out_value in combiner(key, values) or ():
-                write_line(out_key, out_value)
+        for key, pairs_for_key in itertools.groupby(pairs, lambda k_v: k_v[0]):
+            values = (value for _, value in pairs_for_key)
+            for k, v in task(key, values) or ():
+                yield k, v
 
-        if combiner_final:
-            for out_key, out_value in combiner_final() or ():
-                write_line(out_key, out_value)
+        if task_final:
+            for k, v in task_final() or ():
+                yield k, v
 
     def run_spark(self, step_num):
         """Run the Spark code for the given step.
@@ -627,9 +672,9 @@ class MRJob(MRJobLauncher):
         """
         step = self._get_step(step_num, SparkStep)
 
-        if len(self.args) != 2:
+        if len(self.options.args) != 2:
             raise ValueError('Wrong number of args')
-        input_path, output_path = self.args
+        input_path, output_path = self.options.args
 
         spark_method = step.spark
         spark_method(input_path, output_path)
@@ -643,6 +688,8 @@ class MRJob(MRJobLauncher):
         Called from :py:meth:`run`. You'd probably only want to call this
         directly from automated tests.
         """
+        log.warning('--steps is deprecated and going away in v0.7.0')
+
         # json only uses strings, but self.stdout only accepts bytes
         steps_json = json.dumps(self._steps_desc())
         if not isinstance(steps_json, bytes):
@@ -678,16 +725,20 @@ class MRJob(MRJobLauncher):
         - If path is ``-``, read from STDIN.
         - Recursively read all files in a directory
         """
-        paths = self.args or ['-']
+        paths = self.options.args or ['-']
+
         for path in paths:
-            for line in read_input(path, stdin=self.stdin):
-                yield line
+            if path == '-':
+                for line in self.stdin:
+                    yield line
+            else:
+                with open(path, 'rb') as f:
+                    for line in to_lines(decompress(f, path)):
+                        yield line
 
     def _wrap_protocols(self, step_num, step_type):
         """Pick the protocol classes to use for reading and writing
-        for the given step, and wrap them so that bad input and output
-        trigger a counter rather than an exception unless --strict-protocols
-        is set.
+        for the given step.
 
         Returns a tuple of ``(read_lines, write_line)``
 
@@ -704,29 +755,12 @@ class MRJob(MRJobLauncher):
 
         def read_lines():
             for line in self._read_input():
-                try:
-                    key, value = read(line.rstrip(b'\r\n'))
-                    yield key, value
-                except Exception as e:
-                    # the strict_protocols option has to default to None
-                    # because it's used by runners, so treat None as true
-                    if self.options.strict_protocols is not False:
-                        raise
-                    else:
-                        self.increment_counter(
-                            'Undecodable input', e.__class__.__name__)
+                key, value = read(line.rstrip(b'\r\n'))
+                yield key, value
 
         def write_line(key, value):
-            try:
-                self.stdout.write(write(key, value))
-                self.stdout.write(b'\n')
-            except Exception as e:
-                # None counts as true, see above
-                if self.options.strict_protocols is not False:
-                    raise
-                else:
-                    self.increment_counter(
-                        'Unencodable output', e.__class__.__name__)
+            self.stdout.write(write(key, value))
+            self.stdout.write(b'\n')
 
         return read_lines, write_line
 
@@ -743,16 +777,16 @@ class MRJob(MRJobLauncher):
         mapping = {}
         script_step_num = 0
         for i, step in enumerate(steps_desc):
-            if 'mapper' in step:
-                if step['mapper']['type'] == 'script':
-                    k = self._step_key(i, 'mapper')
-                    mapping[k] = script_step_num
-                    script_step_num += 1
-            if 'reducer' in step:
-                if step['reducer']['type'] == 'script':
-                    k = self._step_key(i, 'reducer')
-                    mapping[k] = script_step_num
-                    script_step_num += 1
+
+            if 'mapper' in step and step['mapper']['type'] == 'script':
+                k = self._step_key(i, 'mapper')
+                mapping[k] = script_step_num
+                script_step_num += 1
+
+            if 'reducer' in step and step['reducer']['type'] == 'script':
+                k = self._step_key(i, 'reducer')
+                mapping[k] = script_step_num
+                script_step_num += 1
 
         return mapping
 
@@ -830,55 +864,39 @@ class MRJob(MRJobLauncher):
 
     ### Command-line arguments ###
 
-    def configure_options(self):
+    def configure_args(self):
         """Define arguments for this script. Called from :py:meth:`__init__()`.
 
         Re-define to define custom command-line arguments or pass
         through existing ones::
 
-            def configure_options(self):
-                super(MRYourJob, self).configure_options()
+            def configure_args(self):
+                super(MRYourJob, self).configure_args()
 
-                self.add_passthrough_option(...)
-                self.add_file_option(...)
-                self.pass_through_option(...)
+                self.add_passthru_arg(...)
+                self.add_file_arg(...)
+                self.pass_arg_through(...)
                 ...
         """
+        super(MRJob, self).configure_args()
 
-        super(MRJob, self).configure_options()
-
-        # To run mappers or reducers
-        self._mux_opt_group = OptionGroup(
-            self.option_parser, 'Running specific parts of the job')
-        self.option_parser.add_option_group(self._mux_opt_group)
-
-        _add_step_options(self._mux_opt_group)
-
-    def all_option_groups(self):
-        return super(MRJob, self).all_option_groups() + (self._mux_opt_group,)
+        _add_step_args(self.arg_parser, include_deprecated=True)
 
     def is_task(self):
         """True if this is a mapper, combiner, reducer, or Spark script.
 
-        This is mostly useful inside :py:meth:`load_options`, to disable
-        loading options when we aren't running inside Hadoop.
+        This is mostly useful inside :py:meth:`load_args`, to disable
+        loading args when we aren't running inside Hadoop.
         """
         return (self.options.run_mapper or
                 self.options.run_combiner or
                 self.options.run_reducer or
                 self.options.run_spark)
 
-    def _process_args(self, args):
-        """mrjob.launch takes the first arg as the script path, but mrjob.job
-        uses all args as input files. This method determines the behavior:
-        MRJob uses all args as input files.
-        """
-        self.args = args
-
     def _print_help(self, options):
         """Implement --help --steps"""
         if options.show_steps:
-            _print_help_for_steps()
+            _print_help_for_steps(include_deprecated=self.options.deprecated)
         else:
             super(MRJob, self)._print_help(options)
 
@@ -949,15 +967,29 @@ class MRJob(MRJobLauncher):
     #: See :py:data:`mrjob.protocol` for the full list of protocols.
     OUTPUT_PROTOCOL = JSONProtocol
 
+    def parse_output(self, chunks):
+        """Parse the final output of this MRJob (as a stream of byte chunks)
+        into a stream of ``(key, value)``.
+        """
+        read = self.output_protocol().read
+
+        for line in to_lines(chunks):
+            yield read(line)
+
     def parse_output_line(self, line):
         """
         Parse a line from the final output of this MRJob into
-        ``(key, value)``. Used extensively in tests like this::
+        ``(key, value)``.
 
-            runner.run()
-            for line in runner.stream_output():
-                key, value = mr_job.parse_output_line(line)
+        .. deprecated:: 0.6.0
+
+           Use :py:meth:`parse_output` instead.
         """
+        if not self._warned_about_parse_output_line:
+            log.warning('parse_output_line() is deprecated and will be removed'
+                        ' in v0.7.0; use parse_output() instead.')
+            self._warned_about_parse_output_line = True
+
         return self.output_protocol().read(line)
 
     ### Hadoop Input/Output Formats ###
@@ -1005,7 +1037,7 @@ class MRJob(MRJobLauncher):
     ### Libjars ###
 
     #: Optional list of paths of jar files to run our job with using Hadoop's
-    #: ``-libjar`` option.
+    #: ``-libjars`` option.
     #:
     #: ``~`` and environment variables
     #: in paths be expanded, and relative paths will be interpreted as
@@ -1020,23 +1052,24 @@ class MRJob(MRJobLauncher):
 
     def libjars(self):
         """Optional list of paths of jar files to run our job with using
-        Hadoop's ``-libjar`` option. Normally setting :py:attr:`LIBJARS`
-        is sufficient.
-
-        By default, this combines :option:`libjars` options from the command
-        lines with :py:attr:`LIBJARS`, with command line arguments taking
-        precedence. Paths from :py:attr:`LIBJARS` are interpreted as relative
-        to the the directory containing the script (paths from the
+        Hadoop's ``-libjars`` option. Normally setting :py:attr:`LIBJARS`
+        is sufficient. Paths from :py:attr:`LIBJARS` are interpreted as
+        relative to the the directory containing the script (paths from the
         command-line are relative to the current working directory).
 
         Note that ``~`` and environment variables in paths will always be
         expanded by the job runner (see :mrjob-opt:`libjars`).
 
         .. versionadded:: 0.5.3
+
+        .. versionchanged:: 0.6.6
+
+           re-defining this no longer clobbers command-line
+           ``--libjar`` options.
         """
         script_dir = os.path.dirname(self.mr_job_script())
 
-        paths_from_libjars = []
+        paths = []
 
         # libjar paths will eventually be combined with combine_path_lists,
         # which will expand environment variables. We don't want to assume
@@ -1045,11 +1078,11 @@ class MRJob(MRJobLauncher):
         # prematurely.
         for path in self.LIBJARS or []:
             if os.path.isabs(expand_path(path)):
-                paths_from_libjars.append(path)
+                paths.append(path)
             else:
-                paths_from_libjars.append(os.path.join(script_dir, path))
+                paths.append(os.path.join(script_dir, path))
 
-        return combine_lists(paths_from_libjars, self.options.libjars)
+        return paths
 
     ### Partitioning ###
 
@@ -1064,15 +1097,146 @@ class MRJob(MRJobLauncher):
         """Optional Hadoop partitioner class to use to determine how mapper
         output should be sorted and distributed to reducers.
 
-        By default, returns whatever is passed to :option:`--partitioner`,
-        or if that option isn't used, :py:attr:`PARTITIONER`, or if that
-        isn't set, and :py:attr:`SORT_VALUES` is true, it's set to
-        ``'org.apache.hadoop.mapred.lib.KeyFieldBasedPartitioner'``.
+        By default, returns :py:attr:`PARTITIONER`.
 
         You probably don't need to re-define this; it's just here for
         completeness.
         """
-        return self.options.partitioner or self.PARTITIONER
+        return self.PARTITIONER
+
+    ### Uploading support files ###
+
+    #: Optional list of archives to upload and unpack in the job's working
+    #: directory. These can be URIs or paths on the local filesystem.
+    #:
+    #: Relative paths will be interpreted as relative to the directory
+    #: containing the script (not the current working directory).
+    #
+    #: Environment variables and ``~`` in paths will be expanded.
+    #:
+    #: By default, the directory will have the same name as the archive
+    #: (e.g. ``foo.tar.gz/``). To change the directory's name, append
+    #: ``#<name>``::
+    #:
+    #:     ARCHIVES = ['data/foo.tar.gz#foo']
+    #:
+    #: If you need to dynamically generate a list of files, override
+    #: :py:meth:`archives` instead.
+    #:
+    #: .. versionadded:: 0.6.4
+    ARCHIVES = []
+
+    #: Optional list of directories to upload to the job's working directory.
+    #: These can be URIs or paths on the local filesystem.
+    #:
+    #: Relative paths will be interpreted as relative to the directory
+    #: containing the script (not the current working directory).
+    #
+    #: Environment variables and ``~`` in paths will be expanded.
+    #:
+    #: If you want a directory to be copied with a name other than it's own,
+    #: append ``#<name>`` (e.g. ``data/foo#bar``).
+    #:
+    #: If you need to dynamically generate a list of files, override
+    #: :py:meth:`dirs` instead.
+    #:
+    #: .. versionadded:: 0.6.4
+    DIRS = []
+
+    #: Optional list of files to upload to the job's working directory.
+    #: These can be URIs or paths on the local filesystem.
+    #:
+    #: Relative paths will be interpreted as relative to the directory
+    #: containing the script (not the current working directory).
+    #
+    #: Environment variables and ``~`` in paths will be expanded.
+    #:
+    #: If you want a file to be uploaded to a filename other than it's own,
+    #: append ``#<name>`` (e.g. ``data/foo.json#bar.json``).
+    #:
+    #: If you need to dynamically generate a list of files, override
+    #: :py:meth:`files` instead.
+    #:
+    #: .. versionadded:: 0.6.4
+    FILES = []
+
+    def archives(self):
+        """Like :py:attr:`ARCHIVES`, except that it can return a dynamically
+        generated list of archives to upload and unpack. Overriding
+        this method disables :py:attr:`ARCHIVES`.
+
+        Paths returned by this method are relative to the working directory
+        (not the script). Note that the job runner will *always* expand
+        environment variables and ``~`` in paths returned by this method.
+
+        You do not have to worry about inadvertently disabling ``--archive``;
+        this switch is handled separately.
+
+        .. versionadded:: 0.6.4
+        """
+        return self._upload_attr('ARCHIVES')
+
+    def dirs(self):
+        """Like :py:attr:`DIRS`, except that it can return a dynamically
+        generated list of directories to upload. Overriding
+        this method disables :py:attr:`DIRS`.
+
+        Paths returned by this method are relative to the working directory
+        (not the script). Note that the job runner will *always* expand
+        environment variables and ``~`` in paths returned by this method.
+
+        You do not have to worry about inadvertently disabling ``--dir``;
+        this switch is handled separately.
+
+        .. versionadded:: 0.6.4
+        """
+        return self._upload_attr('DIRS')
+
+    def files(self):
+        """Like :py:attr:`FILES`, except that it can return a dynamically
+        generated list of files to upload. Overriding
+        this method disables :py:attr:`FILES`.
+
+        Paths returned by this method are relative to the working directory
+        (not the script). Note that the job runner will *always* expand
+        environment variables and ``~`` in paths returned by this method.
+
+        You do not have to worry about inadvertently disabling ``--file``;
+        this switch is handled separately.
+
+        .. versionadded:: 0.6.4
+        """
+        return self._upload_attr('FILES')
+
+    def _upload_attr(self, attr_name):
+        """Helper for :py:meth:`archives`, :py:meth:`dirs`, and
+        :py:meth:`files`"""
+        attr_value = getattr(self, attr_name)
+
+        # catch path instead of a list of paths
+        if isinstance(attr_value, string_types):
+            raise TypeError('%s must be a list or other sequence.' % attr_name)
+
+        script_dir = os.path.dirname(self.mr_job_script())
+        paths = []
+
+        for path in attr_value:
+            expanded_path = expand_path(path)
+
+            if os.path.isabs(expanded_path):
+                paths.append(path)
+            else:
+                # relative subdirs are confusing; people will expect them
+                # to appear in a subdir, not the same directory as the script,
+                # but Hadoop doesn't work that way
+                if os.sep in path.rstrip(os.sep) and '#' not in path:
+                    log.warning(
+                        '%s: %s will appear in same directory as job script,'
+                        ' not a subdirectory' % (attr_name, path))
+
+                paths.append(os.path.join(script_dir, path))
+
+        return paths
 
     ### Jobconf ###
 
@@ -1088,74 +1252,14 @@ class MRJob(MRJobLauncher):
 
     def jobconf(self):
         """``-D`` args to pass to hadoop streaming. This should be a map
-        from property name to value.
+        from property name to value. By default, returns :py:attr:`JOBCONF`.
 
-        By default, this combines :option:`jobconf` options from the command
-        lines with :py:attr:`JOBCONF`, with command line arguments taking
-        precedence.
+        .. versionchanged:: 0.6.6
 
-        We also blank out ``mapred.output.key.comparator.class``
-        and ``mapred.text.key.comparator.options`` to prevent interference
-        from :file:`mrjob.conf`.
-
-        :py:attr:`SORT_VALUES` *can* be overridden by :py:attr:`JOBCONF`, the
-        command line, and step-specific ``jobconf`` values.
-
-        For example, if you know your values are numbers, and want to sort
-        them in reverse, you could do::
-
-            SORT_VALUES = True
-
-            JOBCONF = {
-              'mapred.output.key.comparator.class':
-                  'org.apache.hadoop.mapred.lib.KeyFieldBasedComparator',
-              'mapred.text.key.comparator.options': '-k1 -k2nr',
-            }
-
-        If you want to re-define this, it's strongly recommended that do
-        something like this, so as not to inadvertently disable
-        the :option:`jobconf` option::
-
-            def jobconf(self):
-                orig_jobconf = super(MyMRJobClass, self).jobconf()
-                custom_jobconf = ...
-
-                return mrjob.conf.combine_dicts(orig_jobconf, custom_jobconf)
+           re-defining longer clobbers command-line
+           ``--jobconf`` options.
         """
-
-        # deal with various forms of bad behavior by users
-        unfiltered_jobconf = combine_dicts(self.JOBCONF, self.options.jobconf)
-        filtered_jobconf = {}
-
-        def format_hadoop_version(v_float):
-            if v_float >= 1.0:
-                # e.g. 1.0
-                return '%.1f' % v_float
-            else:
-                # e.g. 0.20
-                return '%.2f' % v_float
-
-        for key in unfiltered_jobconf:
-            unfiltered_val = unfiltered_jobconf[key]
-            filtered_val = unfiltered_val
-
-            # boolean values need to be lowercased
-            if isinstance(unfiltered_val, bool):
-                if unfiltered_val:
-                    filtered_val = 'true'
-                else:
-                    filtered_val = 'false'
-
-            # TODO: why would a jobconf variable be named 'hadoop_version'?
-            # hadoop_version should be a string
-            elif (key == 'hadoop_version' and
-                    isinstance(unfiltered_val, float)):
-                log.warning('hadoop_version should be a string, not %s' %
-                            unfiltered_val)
-                filtered_val = format_hadoop_version(unfiltered_val)
-            filtered_jobconf[key] = filtered_val
-
-        return filtered_jobconf
+        return dict(self.JOBCONF)
 
     ### Secondary Sort ###
 
@@ -1176,9 +1280,6 @@ class MRJob(MRJobLauncher):
     #: you're using.
     #:
     #: See :py:meth:`jobconf()` and :py:meth:`partitioner()` for more about
-    #: how this works.
-    #:
-    #: .. versionadded:: 0.4.1
     SORT_VALUES = None
 
     def sort_values(self):
