@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
-# Copyright 2015-2016 Yelp
+# Copyright 2015-2017 Yelp
+# Copyright 2018 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,19 +15,23 @@
 # limitations under the License.
 """Parse "task" logs, which are the syslog and stderr for each individual
 task and typically appear in the userlogs/ directory."""
+from collections import defaultdict
+from logging import getLogger
 import re
 
 from .ids import _add_implied_task_id
 from .ids import _to_job_id
 from .log4j import _parse_hadoop_log4j_records
-from .wrap import _cat_log
+from .wrap import _cat_log_lines
 from .wrap import _ls_logs
+from mrjob import parse
 
+log = getLogger(__name__)
 
 # Match a java exception, possibly preceded by 'PipeMapRed failed!', etc.
 # use this with search()
 _JAVA_TRACEBACK_RE = re.compile(
-    r'$\s+at .*\((.*\.(java|scala):\d+|Native Method)\)$',
+    r'\s+at .*\((.*\.(java|scala):\d+|Native Method)\)$',
     re.MULTILINE)
 
 # Match an error stating that Spark's subprocess has failed (and thus we
@@ -49,9 +54,23 @@ _PRE_YARN_TASK_LOG_PATH_RE = re.compile(
     r'(?P<attempt_num>\d+))/'
     r'(?P<log_type>stderr|syslog)(?P<suffix>\.\w{1,3})?$')
 
+# ignore counters and status (only happens in sim runners, where task stderr
+# is dumped straight to a file).
+
+# convert from bytes regex to text regex
+_COUNTER_RE = re.compile(parse._COUNTER_RE.pattern.decode('ascii'))
+_STATUS_RE = re.compile(parse._STATUS_RE.pattern.decode('ascii'))
 
 # ignore warnings about initializing log4j in task stderr
-_TASK_STDERR_IGNORE_RE = re.compile(r'^log4j:WARN .*$')
+_LOG4J_WARN_RE = re.compile(r'^log4j:WARN .*$')
+
+# also ignore counters and status messages (this only happens in
+# local mode, where there's no real Hadoop to filter them out)
+_TASK_STDERR_IGNORE_RES = [
+    _COUNTER_RE,
+    _STATUS_RE,
+    _LOG4J_WARN_RE,
+]
 
 # this is the start of a Java stacktrace that Hadoop 1 always logs to
 # stderr when tasks fail (see #1430)
@@ -74,7 +93,8 @@ _YARN_TASK_LOG_PATH_RE = re.compile(
     r'(?P<log_type>stderr|stdout|syslog)(?P<suffix>\.\w{1,3})?$')
 
 
-def _ls_task_logs(fs, log_dir_stream, application_id=None, job_id=None):
+def _ls_task_logs(fs, log_dir_stream, application_id=None, job_id=None,
+                  error_attempt_ids=None, attempt_to_container_id=None):
     """Yield matching logs, optionally filtering by application_id
     or job_id.
 
@@ -83,29 +103,16 @@ def _ls_task_logs(fs, log_dir_stream, application_id=None, job_id=None):
     corresponding syslog (stderr logs without a corresponding syslog won't be
     included).
     """
-    stderr_logs = []
-    syslogs = []
-
-    for match in _ls_logs(fs, log_dir_stream, _match_task_log_path,
-                          application_id=application_id,
-                          job_id=job_id):
-        if match['log_type'] == 'stderr':
-            stderr_logs.append(match)
-        elif match['log_type'] == 'syslog':
-            syslogs.append(match)
-
-    key_to_syslog = dict((_log_key(match), match) for match in syslogs)
-
-    for stderr_log in stderr_logs:
-        stderr_log['syslog'] = key_to_syslog.get(_log_key(stderr_log))
-
-    # exclude stderr logs with no syslog
-    stderr_logs_with_syslog = [m for m in stderr_logs if m['syslog']]
-
-    return stderr_logs_with_syslog + syslogs
+    return _ls_task_logs_helper(
+        fs, log_dir_stream, is_spark=False,
+        application_id=application_id, job_id=job_id,
+        error_attempt_ids=error_attempt_ids,
+        attempt_to_container_id=attempt_to_container_id)
 
 
-def _ls_spark_task_logs(fs, log_dir_stream, application_id=None, job_id=None):
+def _ls_spark_task_logs(
+        fs, log_dir_stream, application_id=None, job_id=None,
+        error_attempt_ids=None, attempt_to_container_id=None):
     """Yield matching Spark logs, optionally filtering by application_id
     or job_id.
 
@@ -114,23 +121,92 @@ def _ls_spark_task_logs(fs, log_dir_stream, application_id=None, job_id=None):
     corresponding stdout file; whether we process this depends on the content
     of the stderr file.
     """
-    stderr_logs = []
-    key_to_stdout_log = {}
+    return _ls_task_logs_helper(
+        fs, log_dir_stream, is_spark=True,
+        application_id=application_id, job_id=job_id,
+        error_attempt_ids=error_attempt_ids,
+        attempt_to_container_id=attempt_to_container_id)
 
-    for match in _ls_logs(fs, log_dir_stream, _match_task_log_path,
+
+def _ls_task_logs_helper(fs, log_dir_stream, is_spark,
+                         application_id=None, job_id=None,
+                         error_attempt_ids=None, attempt_to_container_id=None):
+    """Helper for _ls_task_logs() and _ls_spark_task_logs().
+
+    *syslog_type* is the type of the log to pair with stderr logs.
+
+    This is actually a bit weird; on Spark, 'stderr' is the equivalent
+    of syslog in Streaming, and 'stdout' is the equivlend of Streaming's
+    stderr.
+
+    For Streaming, we want stderr logs with corresponding syslogs, and if after
+    listing all task logs we don't find any, syslogs.
+
+    For Spark, we want stderr logs (which are equivalent to Streaming syslogs)
+    with corresponding stdouts, and if after listing all task logs we don't
+    find any, stderr logs without corresponding stdouts.
+    """
+    syslog_type = 'stdout' if is_spark else 'syslog'
+
+    error_attempt_ids = error_attempt_ids or ()
+
+    # figure out subdirs to look for logs in
+    if attempt_to_container_id:
+        # YARN
+        subdirs = [
+            attempt_to_container_id[a] for a in error_attempt_ids
+            if a in attempt_to_container_id]
+    else:
+        subdirs = list(error_attempt_ids)
+
+    # only look in subdirs corresponding to failed attempts
+    if subdirs:
+        log_subdir_stream = ([
+            fs.join(log_dir, subdir)
+            for subdir in subdirs
+            for log_dir in log_dir_list
+        ] for log_dir_list in log_dir_stream)
+    else:
+        log_subdir_stream = log_dir_stream
+
+    key_to_type_to_match = defaultdict(dict)
+
+    # less desirable errors to yield if we don't find the ones we want
+    other_matches = []
+
+    for match in _ls_logs(fs, log_subdir_stream, _match_task_log_path,
                           application_id=application_id,
                           job_id=job_id):
-        if match['log_type'] == 'stderr':
-            stderr_logs.append(match)
-        elif match['log_type'] == 'stdout':
-            key_to_stdout_log[_log_key(match)] = match
 
-    for stderr_log in stderr_logs:
-        stdout_log = key_to_stdout_log.get(_log_key(stderr_log))
-        if stdout_log:
-            stderr_log['stdout'] = stdout_log
+        log_key = _log_key(match)
+        log_type = match['log_type']
 
-    return stderr_logs
+        if log_type not in ('stderr', syslog_type):
+            continue  # don't care
+
+        type_to_match = key_to_type_to_match[log_key]
+
+        if log_type in type_to_match:
+            continue  # already seen
+
+        type_to_match[log_type] = match
+
+        # yield stderrs with syslogs as we find them
+        if 'stderr' in type_to_match and syslog_type in type_to_match:
+            stderr_match = type_to_match['stderr']
+            syslog_match = type_to_match[syslog_type]
+
+            stderr_match[syslog_type] = syslog_match
+
+            yield stderr_match
+
+        if log_type == ('stderr' if is_spark else syslog_type):
+            other_matches.append(match)
+
+    # yield logs that don't have both syslog and stderr
+    for other_match in other_matches:
+        if syslog_type not in other_match:  # already yielded
+            yield other_match
 
 
 def _log_key(match):
@@ -219,7 +295,7 @@ def _interpret_task_logs(fs, matches, partial=True, log_callback=None):
         if stderr_path:
             if log_callback:
                 log_callback(stderr_path)
-            task_error = _parse_task_stderr(_cat_log(fs, stderr_path))
+            task_error = _parse_task_stderr(_cat_log_lines(fs, stderr_path))
 
             if task_error:
                 task_error['path'] = stderr_path
@@ -233,7 +309,7 @@ def _interpret_task_logs(fs, matches, partial=True, log_callback=None):
 
         if log_callback:
             log_callback(syslog_path)
-        syslog_error = _parse_task_syslog(_cat_log(fs, syslog_path))
+        syslog_error = _parse_task_syslog(_cat_log_lines(fs, syslog_path))
         syslogs_parsed.add(syslog_path)
 
         if not syslog_error.get('hadoop_error'):
@@ -303,7 +379,7 @@ def _interpret_spark_task_logs(fs, matches, partial=True, log_callback=None):
         if log_callback:
             log_callback(stderr_path)
         # stderr is Spark's syslog
-        stderr_error = _parse_task_syslog(_cat_log(fs, stderr_path))
+        stderr_error = _parse_task_syslog(_cat_log_lines(fs, stderr_path))
 
         if stderr_error.get('hadoop_error'):
             stderr_error['hadoop_error']['path'] = stderr_path
@@ -318,7 +394,7 @@ def _interpret_spark_task_logs(fs, matches, partial=True, log_callback=None):
             if log_callback:
                 log_callback(stdout_path)
             # the stderr of the application master ends up in "stdout"
-            task_error = _parse_task_stderr(_cat_log(fs, stdout_path))
+            task_error = _parse_task_stderr(_cat_log_lines(fs, stdout_path))
 
             if task_error:
                 task_error['path'] = stdout_path
@@ -357,9 +433,15 @@ def _parse_task_syslog(lines):
         num_lines: (optional) number of lines in split
         start_line: (optional) first line of split (0-indexed)
     """
+    return _parse_task_syslog_records(_parse_hadoop_log4j_records(lines))
+
+
+def _parse_task_syslog_records(records):
+    """Helper for _parse_task_syslog(); takes log4j records rather than
+    lines"""
     result = {}
 
-    for record in _parse_hadoop_log4j_records(lines):
+    for record in records:
         message = record['message']
 
         m = _OPENING_FOR_READING_RE.match(message)
@@ -399,6 +481,8 @@ def _parse_task_syslog(lines):
     return result
 
 
+# TODO: allow filtering of bad lines to happen elsewhere, pass this
+# function numbered lines
 def _parse_task_stderr(lines):
     """Attempt to explain any error in task stderr, be it a Python
     exception or a problem with a setup command (see #1203).
@@ -433,21 +517,24 @@ def _parse_task_stderr(lines):
             else:
                 stack_trace_start_line = None
 
-        # ignore warnings about initializing log4j
-        if _TASK_STDERR_IGNORE_RE.match(line):
+        # ignore warnings about initializing log4j, counters, etc.
+        if any(ir.match(line) for ir in _TASK_STDERR_IGNORE_RES):
             # ignored lines shouldn't count as part of the line range
-            if task_error and 'num_lines' not in task_error:
+            if task_error and task_error.get('num_lines') is None:
                 task_error['num_lines'] = line_num - task_error['start_line']
             continue
         elif not task_error or line.startswith('+ '):
+            # stderr log should only contain counters and status
+            # messages in local mode
             task_error = dict(
                 message=line,
                 start_line=line_num)
         else:
             task_error['message'] += '\n' + line
+            task_error['num_lines'] = None
 
     if task_error:
-        if 'num_lines' not in task_error:
+        if task_error.get('num_lines') is None:
             end_line = stack_trace_start_line or (line_num + 1)
             task_error['num_lines'] = end_line - task_error['start_line']
         return task_error
