@@ -1,6 +1,7 @@
 # Copyright 2012 Yelp and Contributors
 # Copyright 2013 David Marin and Contributors
 # Copyright 2015-2017 Yelp
+# Copyright 2018 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,6 +14,14 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+"""Representations of job steps, to use in your :py:class:`~mrjob.job.MRJob`'s
+:py:meth:`~mrjob.job.MRJob.steps` method.
+
+Because :py:class:`the runner <mrjob.runner.MRJobRunner>` just needs to know
+how to invoke your MRJob script, not how it works insternally, each step
+instance's ``description()`` method produces a simplified, JSON-able
+description of the step, to pass to the runner.
+"""
 import logging
 
 from mrjob.py2 import string_types
@@ -23,7 +32,7 @@ STEP_TYPES = ('jar', 'spark', 'spark_jar', 'spark_script', 'streaming')
 
 # Function names mapping to mapper, reducer, and combiner operations
 _MAPPER_FUNCS = ('mapper', 'mapper_init', 'mapper_final', 'mapper_cmd',
-                 'mapper_pre_filter')
+                 'mapper_pre_filter', 'mapper_raw')
 _COMBINER_FUNCS = ('combiner', 'combiner_init', 'combiner_final',
                    'combiner_cmd', 'combiner_pre_filter')
 _REDUCER_FUNCS = ('reducer', 'reducer_init', 'reducer_final', 'reducer_cmd',
@@ -48,16 +57,21 @@ _SPARK_JAR_STEP_KWARGS = ['args', 'jar', 'main_class', 'spark_args']
 _SPARK_SCRIPT_STEP_KWARGS = ['args', 'script', 'spark_args']
 
 
-#: If passed as an argument to :py:class:`JarStep` or
-#: py:class:`SparkScriptStep`, it'll be replaced with the step's input path(s)
-#: (if there are multiple paths, they'll be joined with commas)
+#: If passed as an argument to :py:class:`JarStep`, :py:class:`SparkJarStep`,
+#: or :py:class:`SparkScriptStep`, it'll be replaced with the step's input
+#: path(s). If there are multiple paths, they'll be joined with commas.
 INPUT = '<input>'
 
-#: If this is passed as passed as an argument to :py:class:`JarStep` or
-#: py:class:`SparkScriptStep`, it'll be replaced
+#: If this is passed as an argument to :py:class:`JarStep`,
+#: :py:class:`SparkJarStep`, or :py:class:`SparkScriptStep`, it'll be replaced
 #: with the step's output path
 OUTPUT = '<output>'
 
+#: If this is passed as an argument to :py:class:`JarStep`,
+#: it'll be replaced with generic hadoop args (-D and -libjars)
+GENERIC_ARGS = '<generic args>'
+
+
 log = logging.getLogger(__name__)
 
 
@@ -80,6 +94,7 @@ class StepFailedException(Exception):
     you may wish to catch it if you
     :ref:`run your job programatically <runners-programmatically>`.
     """
+    _FIELDS = ('reason', 'step_num', 'num_steps', 'step_desc')
 
     def __init__(
             self, reason=None, step_num=None, num_steps=None, step_desc=None):
@@ -113,48 +128,55 @@ class StepFailedException(Exception):
         return '%s(%s)' % (
             self.__class__.__name__,
             ', '.join(('%s=%r' % (k, getattr(self, k))
-                       for k in ('reason', 'step_num',
-                                 'num_steps', 'step_desc')
+                       for k in self._FIELDS
                        if getattr(self, k) is not None)))
 
 
 class MRStep(object):
+    # this docstring excludes mapper_cmd, etc.
     """Represents steps handled by the script containing your job.
 
     Used by :py:meth:`MRJob.steps <mrjob.job.MRJob.steps>`.
     See :ref:`writing-multi-step-jobs` for sample usage.
 
-    Accepts the following keyword arguments:
+    Takes the following keyword arguments: `combiner`, `combiner_cmd`,
+    `combiner_final`, `combiner_init`, `combiner_pre_filter`, `mapper`,
+    `mapper_cmd`, `mapper_final`, `mapper_init`, `mapper_pre_filter`,
+    `mapper_raw`, `reducer`, `reducer_cmd`, `reducer_final`, `reducer_init`,
+    `reducer_pre_filter`. These should be set to ``None`` or a function
+    with the same signature as the corresponding method in
+    :py:class:`~mrjob.job.MRJob`.
 
-    :param mapper: function with same function signature as
-                   :py:meth:`~mrjob.job.MRJob.mapper`, or ``None`` for an
-                   identity mapper.
-    :param reducer: function with same function signature as
-                    :py:meth:`~mrjob.job.MRJob.reducer`, or ``None`` for no
-                    reducer.
-    :param combiner: function with same function signature as
-                     :py:meth:`~mrjob.job.MRJob.combiner`, or ``None`` for no
-                     combiner.
-    :param mapper_init: function with same function signature as
-                        :py:meth:`~mrjob.job.MRJob.mapper_init`, or ``None``
-                        for no initial mapper action.
-    :param mapper_final: function with same function signature as
-                         :py:meth:`~mrjob.job.MRJob.mapper_final`, or ``None``
-                         for no final mapper action.
-    :param reducer_init: function with same function signature as
-                         :py:meth:`~mrjob.job.MRJob.reducer_init`, or ``None``
-                         for no initial reducer action.
-    :param reducer_final: function with same function signature as
-                          :py:meth:`~mrjob.job.MRJob.reducer_final`, or
-                          ``None`` for no final reducer action.
-    :param combiner_init: function with same function signature as
-                          :py:meth:`~mrjob.job.MRJob.combiner_init`, or
-                          ``None`` for no initial combiner action.
-    :param combiner_final: function with same function signature as
-                           :py:meth:`~mrjob.job.MRJob.combiner_final`, or
-                           ``None`` for no final combiner action.
-    :param jobconf: dictionary with custom jobconf arguments to pass to
-                    hadoop.
+    Also accepts `jobconf`, a dictionary with custom jobconf arguments to pass
+    to hadoop.
+
+    A MRStep's description looks like::
+
+        {
+            'type': 'streaming',
+            'mapper': { ... },
+            'combiner': { ... },
+            'reducer': { ... },
+            'jobconf': { ... },  # dict of Hadoop configuration properties
+        }
+
+    At least one of ``mapper``, ``combiner`` and ``reducer`` need be included.
+    ``jobconf`` is completely optional.
+
+    ``mapper``, ``combiner``, and ``reducer`` are either handled by
+    the script containing your job definition, in which case they look like::
+
+        {
+            'type': 'script',
+            'pre_filter': 'grep -v bad', # optional cmd to filter input
+        }
+
+    or they simply run a command, which looks like::
+
+        {
+            'type': 'command',
+            'command': 'cut -f 1-2', # command to run, as a string
+        }
     """
     def __init__(self, **kwargs):
         # limit which keyword args can be specified
@@ -182,18 +204,17 @@ class MRStep(object):
 
         steps.update(kwargs)
 
-        def _prefix_set(prefix):
-            return set(k for k in steps if k.startswith(prefix) and steps[k])
+        def _check_conflict(func, other_funcs):
+            if steps[func]:
+                for other_func in other_funcs:
+                    if steps[other_func] and other_func != func:
+                        raise ValueError("Can't specify both %s and %s" % (
+                            func, other_func))
 
-        def _check_cmd(cmd, prefix_set):
-            if len(prefix_set) > 1 and cmd in prefix_set:
-                prefix_set.remove(cmd)
-                raise ValueError("Can't specify both %s and %s" % (
-                    cmd, prefix_set))
-
-        _check_cmd('mapper_cmd', _prefix_set('mapper'))
-        _check_cmd('combiner_cmd', _prefix_set('combiner'))
-        _check_cmd('reducer_cmd', _prefix_set('reducer'))
+        _check_conflict('mapper_cmd', _MAPPER_FUNCS)
+        _check_conflict('mapper_raw', ('mapper', 'mapper_pre_filter'))
+        _check_conflict('combiner_cmd', _COMBINER_FUNCS)
+        _check_conflict('reducer_cmd', _REDUCER_FUNCS)
 
         self._steps = steps
 
@@ -225,7 +246,7 @@ class MRStep(object):
             return _IDENTITY_REDUCER
         return self._steps[key]
 
-    def _render_substep(self, cmd_key, pre_filter_key=None):
+    def _render_substep(self, cmd_key, pre_filter_key):
         if self._steps[cmd_key]:
             cmd = self._steps[cmd_key]
             if not isinstance(cmd, string_types):
@@ -250,44 +271,8 @@ class MRStep(object):
     def render_reducer(self):
         return self._render_substep('reducer_cmd', 'reducer_pre_filter')
 
-    def description(self, step_num):
-        """Returns a dictionary representation of this step:
-
-        .. code-block:: js
-
-            {
-                'type': 'streaming',
-                'mapper': { ... },
-                'combiner': { ... },
-                'reducer': { ... },
-                'jobconf': dictionary of Hadoop configuration properties
-            }
-
-        ``jobconf`` is optional, and only one of ``mapper``, ``combiner``,
-        and ``reducer`` need be included.
-
-        ``mapper``, ``combiner``, and ``reducer`` are either handled by
-        the script containing your job definition:
-
-        .. code-block:: js
-
-           {
-               'type': 'script',
-               'pre_filter': (optional) cmd to pass input through, as a string
-           }
-
-        or they simply run a command:
-
-        .. code-block:: js
-
-            {
-                'type': 'command',
-                'command': command to run, as a string
-            }
-
-        See :ref:`steps-format` for examples.
-        """
-        substep_descs = {'type': 'streaming'}
+    def description(self, step_num=0):
+        desc = {'type': 'streaming'}
         # Use a mapper if:
         #   - the user writes one
         #   - it is the first step and we don't want to mess up protocols
@@ -295,15 +280,18 @@ class MRStep(object):
         if (step_num == 0 or
                 self.has_explicit_mapper or
                 self.has_explicit_combiner):
-            substep_descs['mapper'] = self.render_mapper()
+            desc['mapper'] = self.render_mapper()
         if self.has_explicit_combiner:
-            substep_descs['combiner'] = self.render_combiner()
+            desc['combiner'] = self.render_combiner()
         if self.has_explicit_reducer:
-            substep_descs['reducer'] = self.render_reducer()
+            desc['reducer'] = self.render_reducer()
+        if self._steps['mapper_raw']:
+            desc['input_manifest'] = True
         # TODO: verify this is a dict, convert booleans to strings
         if self._steps['jobconf']:
-            substep_descs['jobconf'] = self._steps['jobconf']
-        return substep_descs
+            desc['jobconf'] = self._steps['jobconf']
+
+        return desc
 
 
 class _Step(object):
@@ -390,7 +378,7 @@ class _Step(object):
         else:
             return None
 
-    def description(self, other):
+    def description(self, step_num=0):
         """Return a dictionary representation of this step. See
         :ref:`steps-format` for examples."""
         result = dict(
@@ -422,15 +410,21 @@ class JarStep(_Step):
     *jar* can also be passed as a positional argument
 
     See :ref:`non-hadoop-streaming-jar-steps` for sample usage.
-    """
-    # these are deprecated and will be removed in v0.6.0; use
-    # mrjob.step.INPUT and mrjob.step.OUTPUT instead.
-    #
-    # Unfortunately, creating a "class property" to issue a deprecation
-    # warning for these is way more trouble than it's worth (metaclasses, etc.)
-    INPUT = INPUT
-    OUTPUT = OUTPUT
 
+    Sample description of a JarStep::
+
+        {
+            'type': 'jar',
+            'jar': 'binks.jar.jar',
+            'main_class': 'MyMainMan',  # optional
+            'args': ['argh', 'argh']  # optional
+            'jobconf': { ... }  # optional
+        }
+
+    To give your jar access to input files, an empty output directory,
+    configuration properties, and libjars managed by mrjob, you may include
+    :py:data:`INPUT`, :py:data:`OUTPUT`, and :py:data:`GENERIC_ARGS` in *args*.
+    """
     _STEP_TYPE = 'jar'
 
     _STEP_ATTRS = ['args', 'jar', 'jobconf', 'main_class']
@@ -449,6 +443,14 @@ class SparkStep(_Step):
     :param jobconf: (optional) A dictionary of Hadoop properties
     :param spark_args: (optional) an array of arguments to pass to spark-submit
                        (e.g. ``['--executor-memory', '2G']``).
+
+    Sample description of a SparkStep::
+
+        {
+            'type': 'spark',
+            'jobconf': { ... },  # optional
+            'spark_args': ['--executor-memory', '2G'],  # optional
+        }
     """
     _STEP_TYPE = 'spark'
 
@@ -478,6 +480,21 @@ class SparkJarStep(_Step):
                        (e.g. ``['--executor-memory', '2G']``).
 
     *jar* and *main_class* can also be passed as positional arguments
+
+    Sample description of a SparkJarStep::
+
+        {
+            'type': 'spark_jar',
+            'jar': 'binks.jar.jar',
+            'main_class': 'MyMainMan',  # optional
+            'args': ['argh', 'argh'],  # optional
+            'jobconf': { ... },  # optional
+            'spark_args': ['--executor-memory', '2G'],  # optional
+        }
+
+    To give your Spark JAR access to input files and an empty output directory
+    managed by mrjob, you may include :py:data:`INPUT` and :py:data:`OUTPUT`
+    in *args*.
     """
     _STEP_TYPE = 'spark_jar'
 
@@ -504,6 +521,20 @@ class SparkScriptStep(_Step):
                        (e.g. ``['--executor-memory', '2G']``).
 
     *script* can also be passed as a positional argument
+
+    Sample description of a ScriptStep::
+
+       {
+            'type': 'spark_script',
+            'script': 'my_spark_script.py',
+            'args': ['script_arg1', 'script_arg2'],
+            'jobconf': { ... },  # optional
+            'spark_args': ['--executor-memory', '2G'],  # optional
+        }
+
+    To give your Spark script access to input files and an empty output
+    directory managed by mrjob, you may include :py:data:`INPUT` and
+    :py:data:`OUTPUT` in *args*.
     """
     _STEP_TYPE = 'spark_script'
 
@@ -516,3 +547,8 @@ class SparkScriptStep(_Step):
 def _is_spark_step_type(step_type):
     """Does the given step type indicate that it uses Spark?"""
     return step_type.split('_')[0] == 'spark'
+
+
+def _is_pyspark_step_type(step_type):
+    """Does the given step type indicate that it uses Spark and Python?"""
+    return step_type in ('spark', 'spark_script')
