@@ -1,5 +1,6 @@
 # Copyright 2009-2016 Yelp and Contributors
-# Coypright 2017 Yelp
+# Copyright 2017 Yelp
+# Copyright 2018 Yelp
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,18 +18,17 @@
 
 # don't add imports here that aren't part of the standard Python library,
 # since MRJobs need to run in Amazon's generic EMR environment
-import contextlib
 import glob
-import itertools
 import logging
 import os
+import os.path
 import pipes
 import random
 import shlex
 import shutil
 import sys
 import tarfile
-from copy import deepcopy
+from contextlib import contextmanager
 from datetime import timedelta
 from distutils.spawn import find_executable
 from logging import getLogger
@@ -38,16 +38,10 @@ from zipfile import ZIP_STORED
 from zipfile import ZipFile
 from zipfile import is_zipfile
 
-from mrjob.cat import bunzip2_stream
 from mrjob.cat import decompress
-from mrjob.cat import gunzip_stream
 from mrjob.py2 import PY2
 from mrjob.py2 import string_types
 
-# these used to live in util; hush, pyflakes
-bunzip2_stream
-gunzip_stream
-
 log = getLogger(__name__)
 
 
@@ -56,30 +50,6 @@ class NullHandler(logging.Handler):
         pass
 
 
-def bash_wrap(cmd_str):
-    """Escape single quotes in a shell command string and wrap it with ``bash
-    -c '<string>'``.
-
-    This low-tech replacement works because we control the surrounding string
-    and single quotes are the only character in a single-quote string that
-    needs escaping.
-
-    .. deprecated:: 0.5.8
-    """
-    log.warning('bash_wrap() is deprecated and will be removed in v0.6.0')
-    return "bash -c '%s'" % cmd_str.replace("'", "'\\''")
-
-
-def buffer_iterator_to_line_iterator(chunks):
-    """Alias for :py:func:`to_lines()`. Will be removed in v0.6.0.
-
-    .. deprecated:: 0.5.0
-    """
-    log.warning('buffer_iterator_to_line_iterator() has been renamed to'
-                ' to_lines(). This alias will be removed in v0.6.0')
-    return to_lines(chunks)
-
-
 def cmd_line(args):
     """build a command line that works in a shell.
     """
@@ -123,8 +93,8 @@ def log_to_stream(name=None, stream=None, format=None, level=None,
 
     :type name: str
     :param name: name of the logger, or ``None`` for the root logger
-    :type stderr: file object
-    :param stderr:  stream to log to (default is ``sys.stderr``)
+    :type stream: file object
+    :param stream:  stream to log to (default is ``sys.stderr``)
     :type format: str
     :param format: log message format (default is '%(message)s')
     :param level: log level to use
@@ -155,6 +125,8 @@ def parse_and_save_options(option_parser, args):
     in *args* that correspond to that *dest*.
 
     This won't modify *option_parser*.
+
+    .. deprecated:: 0.6.0
     """
     arg_map = {}
 
@@ -204,10 +176,16 @@ def read_file(path, fileobj=None, yields_lines=True, cleanup=None):
                     is omitted, we ``open(path)``.
     :param yields_lines: Does iterating over *fileobj* yield lines (like
                          file objects are supposed to)? If not, set this to
-                         ``False`` (useful for :py:class:`boto.s3.Key`)
+                         ``False`` (useful for objects that correspond
+                         to objects on cluster filesystems)
     :param cleanup: Optional callback to call with no arguments when EOF is
                     reached or an exception is thrown.
+
+    .. deprecated:: 0.6.0
     """
+    log.warning('read_file() is deprecated and will be removed in v0.7.0.'
+                ' Try mrjob.cat.decompress() and mrjob.util.to_lines()')
+
     # sometimes values declared in the ``try`` block aren't accessible from the
     # ``finally`` block. not sure why.
     f = None
@@ -248,7 +226,19 @@ def read_input(path, stdin=None):
 
     You can redefine *stdin* for ease of testing. *stdin* can actually be
     any iterable that yields lines (e.g. a list).
+
+    .. deprecated:: 0.6.0
     """
+    log.warning('read_input() is deprecated and will be removed in v0.7.0.'
+                ' Try mrjob.cat.decompress() and mrjob.util.to_lines()')
+
+    for line in _read_input(path, stdin=stdin):
+        yield line
+
+
+def _read_input(path, stdin=None):
+    """Helper function for _read_input() (to avoid getting recursive
+    deprecation warnings)"""
     if stdin is None:
         stdin = sys.stdin
 
@@ -321,7 +311,7 @@ def safeeval(expr, globals=None, locals=None):
     return eval(expr, safe_globals, locals)
 
 
-@contextlib.contextmanager
+@contextmanager
 def save_current_environment():
     """ Context manager that saves os.environ and loads
         it back again after execution
@@ -336,7 +326,7 @@ def save_current_environment():
         os.environ.update(original_environ)
 
 
-@contextlib.contextmanager
+@contextmanager
 def save_cwd():
     """Context manager that saves the current working directory,
     and chdir's back to it after execution."""
@@ -349,6 +339,18 @@ def save_cwd():
         os.chdir(original_cwd)
 
 
+@contextmanager
+def save_sys_path():
+    """Context manager that saves sys.path and restores it after execution."""
+    original_sys_path = list(sys.path)
+
+    try:
+        yield
+
+    finally:
+        sys.path = original_sys_path
+
+
 def shlex_split(s):
     """Wrapper around shlex.split(), but convert to str if Python version <
     2.7.3 when unicode support was added.
@@ -367,52 +369,32 @@ def strip_microseconds(delta):
     return timedelta(delta.days, delta.seconds)
 
 
-def tar_and_gzip(dir, out_path, filter=None, prefix=''):
-    """Tar and gzip the given *dir* to a tarball at *out_path*.
-
-    If we encounter symlinks, include the actual file, not the symlink.
-
-    :type dir: str
-    :param dir: dir to tar up
-    :type out_path: str
-    :param out_path: where to write the tarball too
-    :param filter: if defined, a function that takes paths (relative to *dir*
-                   and returns ``True`` if we should keep them
-    :type prefix: str
-    :param prefix: subdirectory inside the tarball to put everything into (e.g.
-                   ``'mrjob'``)
-    """
-    # might move this to tests/; we use it a few places there
-    log.warning('tar_and_gzip() is deprecated and will be removed in v0.6.0')
-
-    if not os.path.isdir(dir):
-        raise IOError('Not a directory: %r' % (dir,))
-
-    if not filter:
-        filter = lambda path: True
-
-    tar_gz = tarfile.open(out_path, mode='w:gz')
-
-    for dirpath, dirnames, filenames in os.walk(dir, followlinks=True):
-        for filename in filenames:
-            path = os.path.join(dirpath, filename)
-            # janky version of os.path.relpath() (Python 2.6):
-            rel_path = path[len(os.path.join(dir, '')):]
-            if filter(rel_path):
-                # copy over real files, not symlinks
-                real_path = os.path.realpath(path)
-                path_in_tar_gz = os.path.join(prefix, rel_path)
-                tar_gz.add(real_path, arcname=path_in_tar_gz, recursive=False)
-
-    tar_gz.close()
-
-
 def to_lines(chunks):
     """Take in data as a sequence of bytes, and yield it, one line at a time.
 
     Only breaks lines on ``\\n`` (not ``\\r``), and does not add
     a trailing newline.
 
+    For efficiency, passes through anything with a ``readline()`` attribute.
+    """
+    # hopefully this is good enough for anything mrjob will encounter
+    if hasattr(chunks, 'readline'):
+        return chunks
+    else:
+        return _to_lines(chunks)
+
+
+def _to_lines(chunks):
+    """Take in data as a sequence of bytes, and yield it, one line at a time.
+
+    Only breaks lines on ``\\n`` (not ``\\r``), and does not add
+    a trailing newline.
+
+    Exception: if we encounter an empty bytestring ``b''``, immediately yield
+    what we have so far rather than joining it to the next chunk. This allows
+    us to handle bytes from multiple files without joining the end of one
+    file to the beginning of the next one.
+
     Optimizes for:
 
     * chunks bigger than lines (e.g. reading test files)
@@ -422,6 +404,14 @@ def to_lines(chunks):
     leftovers = []
 
     for chunk in chunks:
+        # special case for b'' standing for EOF
+        if chunk == b'':
+            if leftovers:
+                yield b''.join(leftovers)
+                leftovers = []
+
+            continue
+
         start = 0
 
         while start < len(chunk):
@@ -471,10 +461,10 @@ def unarchive(archive_path, dest):
     within zip files can be deflated or stored.
     """
     if tarfile.is_tarfile(archive_path):
-        with contextlib.closing(tarfile.open(archive_path, 'r')) as archive:
+        with tarfile.open(archive_path, 'r') as archive:
             archive.extractall(dest)
     elif is_zipfile(archive_path):
-        with contextlib.closing(ZipFile(archive_path, 'r')) as archive:
+        with ZipFile(archive_path, 'r') as archive:
             for name in archive.namelist():
                 # the zip spec specifies that front slashes are always
                 # used as directory separators
@@ -502,6 +492,7 @@ def which(cmd, path=None):
     as a keyword argument.
     """
     if hasattr(shutil, 'which'):
+        # added in Python 3.3
         return shutil.which(cmd, path=path)
     elif path is None and os.environ.get('PATH') is None:
         # find_executable() errors if neither path nor $PATH is set
@@ -531,230 +522,20 @@ def zip_dir(dir, out_path, filter=None, prefix=''):
     if not filter:
         filter = lambda path: True
 
-    try:
-        zip_file = ZipFile(out_path, mode='w', compression=ZIP_DEFLATED)
-    except RuntimeError:  # zlib not available
-        zip_file = ZipFile(out_path, mode='w', compression=ZIP_STORED)
+    def create_zip_file():
+        try:
+            return ZipFile(out_path, mode='w', compression=ZIP_DEFLATED)
+        except RuntimeError:  # zlib not available
+            return ZipFile(out_path, mode='w', compression=ZIP_STORED)
 
-    for dirpath, dirnames, filenames in os.walk(dir, followlinks=True):
-        for filename in filenames:
-            path = os.path.join(dirpath, filename)
-            # janky version of os.path.relpath() (Python 2.6):
-            rel_path = path[len(os.path.join(dir, '')):]
-            if filter(rel_path):
-                # copy over real files, not symlinks
-                real_path = os.path.realpath(path)
-                path_in_zip_file = os.path.join(prefix, rel_path)
-                zip_file.write(real_path, arcname=path_in_zip_file)
+    with create_zip_file() as zip_file:
+        for dirpath, dirnames, filenames in os.walk(dir, followlinks=True):
+            for filename in filenames:
+                path = os.path.join(dirpath, filename)
+                rel_path = os.path.relpath(path, dir)
 
-    zip_file.close()
-
-
-### deprecated code, to remove in v0.6.0 ###
-
-def args_for_opt_dest_subset(option_parser, args, dests=None):
-    """For the given :py:class:`OptionParser` and list of command line
-    arguments *args*, yield values in *args* that correspond to option
-    destinations in the set of strings *dests*. If *dests* is None, return
-    *args* as parsed by :py:class:`OptionParser`.
-
-    .. deprecated:: 0.5.8
-    """
-    log.warning('args_for_opt_dest_subset() is deprecated and will be'
-                ' removed in v0.6.0')
-
-    for dest, value in _args_for_opt_dest_subset(option_parser, args, dests):
-        yield value
-
-
-def _args_for_opt_dest_subset(option_parser, args, dests=None):
-    """See docs for :py:func:`args_for_opt_dest_subset()`. This function allows
-    us to write a compatibility wrapper for the old API
-    (:py:func:`parse_and_save_options()`).
-    """
-    values = deepcopy(option_parser.get_default_values())
-    rargs = [x for x in args]
-    option_parser.rargs = rargs
-    while rargs:
-        arg = rargs[0]
-        if arg == '--':
-            del rargs[0]
-            return
-        elif arg[0:2] == '--':
-            for item in _process_long_opt(option_parser, rargs, values, dests):
-                yield item
-        elif arg[:1] == '-' and len(arg) > 1:
-            for item in _process_short_opts(option_parser, rargs, values,
-                                            dests):
-                yield item
-        else:
-            del rargs[0]
-
-
-def _process_long_opt(option_parser, rargs, values, dests):
-    """Mimic function of the same name in ``OptionParser``, capturing the
-    arguments consumed in *arg_map*
-    """
-    arg = rargs.pop(0)
-
-    # Value explicitly attached to arg?  Pretend it's the next
-    # argument.
-    if "=" in arg:
-        (opt, next_arg) = arg.split("=", 1)
-        rargs.insert(0, next_arg)
-    else:
-        opt = arg
-
-    opt = option_parser._match_long_opt(opt)
-    option = option_parser._long_opt[opt]
-
-    # Store the 'before' value of *rargs*
-    rargs_before_processing = [x for x in rargs]
-
-    if option.takes_value():
-        nargs = option.nargs
-        if nargs == 1:
-            value = rargs.pop(0)
-        else:
-            value = tuple(rargs[0:nargs])
-            del rargs[0:nargs]
-    else:
-        value = None
-
-    option.process(opt, value, values, option_parser)
-
-    if dests is None or option.dest in dests:
-        # Measure rargs before and after processing. Yield difference.
-        length_difference = len(rargs_before_processing) - len(rargs)
-        for item in [opt] + rargs_before_processing[:length_difference]:
-            yield option.dest, item
-
-
-def _process_short_opts(option_parser, rargs, values, dests):
-    """Mimic function of the same name in ``OptionParser``, capturing the
-    arguments consumed in *arg_map*
-    """
-    arg = rargs.pop(0)
-    stop = False
-    i = 1
-    for ch in arg[1:]:
-        opt = "-" + ch
-        option = option_parser._short_opt.get(opt)
-        i += 1                      # we have consumed a character
-
-        # Store the 'before' value of *rargs*
-        rargs_before_processing = [x for x in rargs]
-
-        # We won't see a difference in rargs for things like '-pJSON', so
-        # handle that edge case explicitly.
-        args_from_smashed_short_opt = []
-
-        if option.takes_value():
-            # Any characters left in arg?  Pretend they're the
-            # next arg, and stop consuming characters of arg.
-            if i < len(arg):
-                rargs.insert(0, arg[i:])
-                args_from_smashed_short_opt.append(arg[i:])
-                stop = True
-
-            nargs = option.nargs
-            if nargs == 1:
-                value = rargs.pop(0)
-            else:
-                value = tuple(rargs[0:nargs])
-                del rargs[0:nargs]
-
-        else:                       # option doesn't take a value
-            value = None
-
-        option.process(opt, value, values, option_parser)
-
-        if dests is None or option.dest in dests:
-            # Measure rargs before and after processing. Yield difference.
-            length_difference = len(rargs_before_processing) - len(rargs)
-            for item in ([opt] + args_from_smashed_short_opt +
-                         rargs_before_processing[:length_difference]):
-                yield option.dest, item
-
-        if stop:
-            break
-
-
-def populate_option_groups_with_options(assignments, indexed_options):
-    """Given a dictionary mapping :py:class:`OptionGroup` and
-    :py:class:`OptionParser` objects to a list of strings represention option
-    dests, populate the objects with options from ``indexed_options``
-    (generated by :py:func:`scrape_options_and_index_by_dest`) in alphabetical
-    order by long option name. This function primarily exists to serve
-    :py:func:`scrape_options_into_new_groups`.
-
-    :type assignments: dict of the form ``{my_option_parser: ('verbose',
-                       'help', ...), my_option_group: (...)}``
-    :param assignments: specification of which parsers/groups should get which
-                        options
-    :type indexed_options: dict generated by
-                           :py:func:`util.scrape_options_and_index_by_dest`
-    :param indexed_options: options to use when populating the parsers/groups
-    """
-    log.warning('populate_option_groups_with_options() is deprecated and'
-                ' will be removed in v0.6.0')
-
-    for opt_group, opt_dest_list in assignments.items():
-        new_options = []
-        for option_dest in assignments[opt_group]:
-            for option in indexed_options[option_dest]:
-                new_options.append(option)
-        # New options must be added using add_options() or they will not be
-        # allowed by the parser on the command line
-        opt_group.add_options(new_options)
-        # Sort alphabetically for help
-        opt_group.option_list = sorted(opt_group.option_list,
-                                       key=lambda item: item.get_opt_string())
-
-
-def scrape_options_and_index_by_dest(*parsers_and_groups):
-    """Scrapes ``optparse`` options from :py:class:`OptionParser` and
-    :py:class:`OptionGroup` objects and builds a dictionary of
-    ``dest_var: [option1, option2, ...]``. This function primarily exists to
-    serve :py:func:`scrape_options_into_new_groups`.
-
-    An example return value: ``{'verbose': [<verbose_on_option>,
-    <verbose_off_option>], 'files': [<file_append_option>]}``
-
-    :type parsers_and_groups: :py:class:`OptionParser` or
-                              :py:class:`OptionGroup`
-    :param parsers_and_groups: Parsers and groups to scrape option objects from
-
-    :return: dict of the form ``{dest_var: [option1, option2, ...], ...}``
-    """
-    log.warning('scrape_options_and_index_by_dest() is deprecated'
-                ' and will be removed in v0.6.0')
-
-    all_options = {}
-    job_option_lists = [g.option_list for g in parsers_and_groups]
-    for option in itertools.chain(*job_option_lists):
-        other_options = all_options.get(option.dest, [])
-        other_options.append(option)
-        all_options[option.dest] = other_options
-    return all_options
-
-
-def scrape_options_into_new_groups(source_groups, assignments):
-    """Puts options from the :py:class:`OptionParser` and
-    :py:class:`OptionGroup` objects in `source_groups` into the keys of
-    `assignments` according to the values of `assignments`. An example:
-
-    :type source_groups: list of :py:class:`OptionParser` and
-                         :py:class:`OptionGroup` objects
-    :param source_groups: parsers/groups to scrape options from
-    :type assignments: dict with keys that are :py:class:`OptionParser` and
-                       :py:class:`OptionGroup` objects and values that are
-                       lists of strings
-    :param assignments: map empty parsers/groups to lists of destination names
-                        that they should contain options for
-    """
-    log.warning('scrape_options_into_new_groups() is deprecated'
-                ' and will be removed in v0.6.0')
-
-    all_options = scrape_options_and_index_by_dest(*source_groups)
-    return populate_option_groups_with_options(assignments, all_options)
+                if filter(rel_path):
+                    # copy over real files, not symlinks
+                    real_path = os.path.realpath(path)
+                    path_in_zip_file = os.path.join(prefix, rel_path)
+                    zip_file.write(real_path, arcname=path_in_zip_file)
